/*
 * BongoDB, an embedded document-based engine
 * Copyright (C) 2015 by Javier Sancho Fernandez <jsf at jsancho dot org>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "bongodb.h"

#include <stdlib.h>
#include <string.h>


/* BongoDB Client */

bongo_client_t *
bongo_client_new (void)
{
  bongo_client_t *client;

  client = malloc (sizeof (bongo_client_t));
  client->n_databases = 0;
  client->databases = NULL;
  
  return client;
}

bongo_database_node_t *
bongo_database_node_new (bongo_client_t client,
			 const char *db_name)
{
  bongo_database_node_t *database_node;

  if (client == NULL)
    return NULL;

  database_node = malloc (sizeof (bongo_database_node_t));
  database_node->db_name = db_name;
  database_node->n_collections = 0;
  database_node->collections = NULL;
  database_node->next = (struct bongo_database_node_t *) client->databases;
  
  client->databases = (bongo_database_node_t *) database_node;
  client->n_databases++;

  return database_node;
}

bongo_collection_node_t *
bongo_collection_node_new (bongo_database_node_t *database_node,
			   const char *col_name)
{
  bongo_collection_node_t *collection_node;

  if (database_node == NULL)
    return NULL;

  collection_node = malloc (sizeof (bongo_collection_node_t));
  collection_node->col_name = col_name;
  collection_node->next = (struct bongo_collection_node_t *) database_node->collections;

  database_node->collections = (bongo_collection_node_t *) collection_node;
  database_node->n_collections++;

  return collection_node;
}

bongo_database_node_t *
bongo_database_node_search (bongo_client_t *client,
			    const char *db_name)
{
  bongo_database_node_t *database_node;

  if (client == NULL)
    return NULL;

  database_node = client->databases;
  while (database_node != NULL && strcmp (database_node->db_name, db_name) != 0)
    database_node = (bongo_database_node_t *) database_node->next;

  return database_node;
}

bongo_collection_node_t *
bongo_collection_node_search (bongo_database_node_t *database_node,
			      const char *col_name)
{
  bongo_collection_node_t *collection_node;

  if (database_node == NULL)
    return NULL;

  collection_node = database_node->collections;
  while (collection_node != NULL && strcmp (collection_node->col_name, col_name) != 0)
    collection_node = (bongo_collection_node_t *) collection_node->next;

  return collection_node;
}
  
void
bongo_collection_node_destroy (bongo_collection_node_t *collection_node)
{
  free (collection_node);
}

void
bongo_database_node_destroy (bongo_database_node_t *database_node)
{
  bongo_collection_node_t *collection_node;

  collection_node = database_node->collections;
  while (collection_node != NULL)
    {
      database_node->collections = (bongo_collection_node_t *) collection_node->next;
      collection_node_destroy (collection_node);
      collection_node = database_node->collections;
    }
  
  free (database_node);
}

void
bongo_client_destroy (bongo_client_t *client)
{
  bongo_database_node_t *database_node;

  database_node = client->databases;
  while (database_node != NULL)
    {
      client->databases = (bongo_database_node_t *) database_node->next;
      database_node_destroy (database_node);
      database_node = client->databases;
    }

  free (client);
}


/* BongoDB Database */

bongo_database_t *
bongo_client_get_database (bongo_client_t *client,
			   const char *db_name)
{
  bongo_database_node_t *database_node;
  bongo_database_t *database;

  if (client == NULL)
    return NULL;

  database_node = bongo_database_node_search(client, db_name);
  if (database_node == NULL)
    database_node = bongo_database_node_new(client, db_name);

  database = malloc(sizeof(bongo_database_t));
  database->database_node = database_node;
  
  return database;
}

void
bongo_database_destroy(bongo_database_t *database)
{
  free(database);
}


/* BongoDB Collection */

bongo_collection_t *
bongo_client_get_collection(bongo_client_t *client,
			    const char *db_name,
			    const char *col_name)
{
  bongo_database_t *database;
  bongo_collection_node_t *collection_node;
  bongo_collection_t *collection;

  if (client == NULL)
    return NULL;

  database = bongo_client_get_database(client, db_name);

  collection_node = bongo_collection_node_search(database->database_node, col_name);
  if (collection_node == NULL)
    collection_node = bongo_collection_node_new(database->database_node, col_name);

  bongo_database_destroy(database);

  collection = malloc(sizeof(bongo_collection_t));
  collection->collection_node = collection_node;

  return collection;
}

bool
bongo_collection_insert (bongo_collection_t *collection,
			 bongo_insert_flags_t flags,
			 const char *document,
			 const bongo_write_concern_t *write_concern,
			 bongo_error_t *error)
{
}

void
bongo_collection_destroy(bongo_collection_t *collection)
{
  free(collection);
}
