;;; BongoDB, an embedded document-based engine
;;; Copyright (C) 2016 by Javier Sancho Fernandez <jsf at jsancho dot org>
;;;
;;; This program is free software: you can redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation, either version 3 of the License, or
;;; (at your option) any later version.
;;;
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (bongodb)
  #:use-module (srfi srfi-9)
  #:use-module (srfi srfi-9 gnu)
  #:use-module (srfi srfi-69)
  #:export (make-collection
	    collection?
	    count
	    insert
	    find
	    update
	    $eq
	    $exists
	    $not
	    $and
	    $set))


;;; Collection Definition

(define-record-type collection
  (make-collection-record table)
  collection?
  (table get-table))

(define (make-collection)
  (make-collection-record (make-hash-table)))

(define (count col)
  (hash-table-size (get-table col)))

(set-record-type-printer! collection
  (lambda (record port)
    (format port
	    "#<collection with ~a documents>"
	    (count record))))


;;; Working with documents

(define (format-document document)
  "Return an alist document ready to store in a collection"
  (document-with-id
   (cond ((hash-table? document)
	  (hash-table->alist document))
	 (else
	  document))))

(define (document-with-id document)
  "Return always a document with an id"
  (or (and (assoc '_id document)
	   document)
      (assoc-set! document '_id (gensym))))

(define (insert col . documents)
  "Insert documents into the collection and return document ids"
  (cond ((null? documents)
	 '())
	(else
	 (let* ((document (format-document (car documents)))
		(docid (assoc-ref document '_id)))
	   (hash-table-set! (get-table col) docid document)
	   (cons docid (apply insert (cons col (cdr documents))))))))

(define (find col filter)
  "Query the collection and return the documents that match with filter"
  (hash-table-fold
   (get-table col)
   (lambda (key document result)
     (cond ((match-document? document filter)
	    (cons document result))
	   (else
	    result)))
   '()))

(define (match-document? document filter)
  "Try to match the given document with a given filter"
  (cond ((equal? filter #t)
	 #t)
	(else
	 (filter document))))

(define (update col filter . changes)
  "Update selected documents using the given filter"
  (let ((documents (find col filter)))
    (apply insert (cons col (map (lambda (doc) (update-document doc changes)) documents)))))

(define (update-document document changes)
  "Update a document with the appropiate changes"
  (cond ((null? changes)
	 document)
	(else
	 (update-document ((car changes) document) (cdr changes)))))


;;; Queries

(define ($eq field value)
  (lambda (document)
    (let ((stored (assoc field document)))
      (and stored
	   (equal? (cdr stored) value)))))

(define ($exists field)
  (lambda (document)
    (and (assoc field document)
	 #t)))

(define ($not expr)
  (lambda (document)
    (not (expr document))))

(define-syntax $and
  (syntax-rules ()
    ((_ expr ...)
     (lambda (document)
       (and (expr document)
	    ...)))))


;;; Updates

(define ($set field value)
  (lambda (document)
    (assoc-set! document field value)))
